/***************************************************************************
	copyright            : (C) 2007 by Marvin Schmidt
	email                : Thund3r@Virtual-Highway.de
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it  under the terms of the GNU Lesser General Public License version  *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 ***************************************************************************/

#include <getopt.h>

#include <iostream>
#include <fstream>
#include <string>
#include <boost/regex.hpp>
using namespace std;

#include <taglib.h>
#include <mpegfile.h>
#include <id3v2tag.h>
#include <synchronisedlyricsframe.h>

#define VERSION "1.1.0"

TagLib::ID3v2::LyricArray lyrics;

void printUsage(char *sName)
{
	cout << "Usage: " << sName << " [OPTION] [INPUT FILE] [OUTPUT FILE]" << endl;
	cout << "Add lyrics from LRC files to the ID3v2 tag of a MPEG file using the synchronised lyrics/text frame" << endl;
	cout << endl;
	cout << "  -s,  --lrc2sylt           Parses input file (LRC) for lyrics and writes them to output file (MPEG)" << endl;
	cout << "  -h,  --help               Display this help and exit" << endl;
}

void printVersion(char *sName)
{
	cout << sName << " " << VERSION << endl;
	cout << "Using TagLib " << TAGLIB_MAJOR_VERSION << "." << TAGLIB_MINOR_VERSION << "." << TAGLIB_PATCH_VERSION << endl;
	cout << endl;
	cout << "This program currently only allows you to add lyrics from LRC files (simple or enhanced) to MPEG files" << endl;
}

long match2Msecs(basic_string<char> mins, basic_string<char> secs, basic_string<char> msecs)
{
	return atoi(mins.data()) * 60 * 1000 + atoi(secs.data()) * 1000 + atoi(msecs.data());
}

int parseLRCFile(char *lrcFilename)
{
	ifstream myfile(lrcFilename);
	if (!myfile.is_open())
	{
		cerr << "Unable to open file" << endl;
		return -1;
	}

	// match lines
	boost::cmatch lmatch;
	boost::regex e("\\[([0-9]+):([0-9]{2}).?([0-9]{1,3})?\\] *(.+)?"); // e.g [01:23.047] some text...
	string line;
	while (getline(myfile, line))
	{
		if (boost::regex_match(line.data(), lmatch, e))
		{
			TagLib::ID3v2::SyllableMap syllables;
			string lyricLine(lmatch[4]);

			boost::regex f("(<([0-9]+):([0-9]{2}).?([0-9]{1,3})?>(.*?))+");		// e.g. [01:56.70]<01:57.15>I <01:57.50>kept <01:57.90>my <01:58.25>pro<01:58.90>mi<01:59.95>se
			boost::regex g("(.+?)(<([0-9]+):([0-9]{2}).?([0-9]{1,3})?>(.*?))+");	// e.g. [02:00.35]Don't <02:00.75>keep <02:01.15>your <02:01.40>di<02:01.60>s<02:01.75>tan<02:02.95>ce
			boost::smatch what;

			string syllable;
			if (boost::regex_match(lyricLine, what, f, boost::match_extra))
			{
				for (unsigned i = 0; i < what.captures(1).size(); i++)
				{
					syllable = what.captures(5)[i];
					if (syllable.empty())
						syllable = " ";
					syllables.insert(match2Msecs(what.captures(2)[i], what.captures(3)[i], what.captures(4)[i]), TagLib::String(syllable, TagLib::String::UTF8));
				}
			}
			else if (boost::regex_match(lyricLine, what, g, boost::match_extra))
			{
				for(unsigned i = 0; i < what.captures(2).size(); i++)
				{
					syllable = what.captures(6)[i];
					if (syllable.empty())
						syllable = " ";
					syllables.insert(match2Msecs(what.captures(3)[i], what.captures(4)[i], what.captures(5)[i]), TagLib::String(syllable, TagLib::String::UTF8));
				}
			}
			else
			{
				if (lyricLine.empty())
					lyricLine = " ";
				syllables.insert(match2Msecs(lmatch[1], lmatch[2], lmatch[3]), TagLib::String(lyricLine, TagLib::String::UTF8));
			}
			lyrics.append( syllables );
		}
	}
	myfile.close();

	return lyrics.size();
}


int lrc2Sylt(char *lrcFilename, char *mpegFilename)
{
	cout << "* Parsing LRC file \"" << lrcFilename << "\"" << endl;
	if (parseLRCFile(lrcFilename) < 0)
		return -1;

	cout << "* Embedding lyrics into MPEG file \"" << mpegFilename << "\"" << endl;
	TagLib::MPEG::File f(mpegFilename);
	TagLib::ID3v2::Tag *tag = f.ID3v2Tag(true);
	TagLib::ID3v2::FrameList l = tag->frameListMap()["SYLT"];
	if (!l.isEmpty())
	{
		cout << "* SYLT Frame(s) found -> removing" << endl;
		for (TagLib::ID3v2::FrameList::Iterator it = l.begin(); it != l.end(); it++)
			tag->removeFrame(*it);
	}

	cout << "* Adding new SYLT Frame" << endl;
	TagLib::ID3v2::SynchronisedLyricsFrame *frame = new TagLib::ID3v2::SynchronisedLyricsFrame(TagLib::String::UTF8);
	frame->setLyrics(lyrics);
	tag->addFrame(frame);

	cout << "* Saving ID3v2 tag" << endl;
	f.save(TagLib::MPEG::File::ID3v2);

	return 0;
}

int main(int argc, char *argv[])
{
	int iOpt;
	while (true)
	{
		static struct option long_options[] =
		{
			{ "help",		no_argument, 0, 'h' },
			{ "version",	no_argument, 0, 'v' },
			{ "lrc2sylt",	required_argument, 0, 's' },
			{ 0, 0, 0, 0 }
		};
		int option_index = 0;
		iOpt = getopt_long(argc, argv, "hvs:", long_options, &option_index);

		if (iOpt == -1)
			iOpt = '?';

		switch (iOpt)
		{
		case '?':
		case 'h':
			printUsage(argv[0]);
			return 0;

		case 'v':
			printVersion(argv[0]);
			return 0;

		case 's':
			if (optind == argc)
			{
				cerr << "No file to work on." << endl;
				exit(1);
			}
			/*
			for ( = optind; (unsigned int) nIndex < argc; nIndex++)
				if (stat(argv[nIndex], &filestat))
				{
					std::cerr << "Couldn't stat file '" << argv[nIndex] << "'\n";
					break;
				}*/

			lrc2Sylt(optarg, argv[optind]);
			return 0;

		default:
			cerr << "This isn't supposed to happen" << endl;
			return -1;

		}
	}
}
